// Copyright 2016 Orus.io. All rights reserved.

package control

import (
	"context"
	"encoding/json"
	"fmt"
	"io"
	math_bits "math/bits"
	"sync"

	yaml "gopkg.in/yaml.v2"
	"xbus.io/go-xbus/v3/api"
	"xbus.io/go-xbus/v3/api/control"
)

type processServer struct {
	*Service
}

func (s processServer) Query(
	ctx context.Context, req control.ProcessQueryRequest,
	send func(control.ProcessQueryReply),
) error {
	s.TaskAdd()
	defer s.TaskDone()

	pList := s.o.ProcessStorage.Query(req.Filter)
	select {
	case <-ctx.Done():
		return nil
	default:
	}
	var (
		maxSize = 1024 * 512
		size    = 0
		batch   []api.Process
	)
	for _, p := range pList {
		state := s.o.ProcessStorage.GetState(p.ID)
		b, err := json.Marshal(state)
		if err != nil {
			panic(err)
		}
		createdAt := p.CreatedAt
		entry := api.Process{
			ID:                p.ID,
			TriggerEmitterID:  p.TriggerEmitterID,
			TriggerEnvelopeID: p.TriggerEnvelopeID,
			CreatedAt:         &createdAt,
			PipelineID:        p.PipelineID,
			GroupID:           p.GroupID,
			ReplayOf:          p.ReplayOf,
			Status:            state.Status,
			State:             string(b),
		}
		select {
		case <-ctx.Done():
			return nil
		default:
		}
		entrySize := entry.Size()
		entrySizeDelta := entrySize + 1 + sovControl(uint64(entrySize))
		if size+entrySizeDelta > maxSize {
			send(control.ProcessQueryReply{ProcessList: batch})
			batch = batch[0:0]
			size = 0
		}
		select {
		case <-ctx.Done():
			return nil
		default:
		}
		batch = append(batch, entry)
		size += entrySizeDelta
	}
	if len(batch) != 0 {
		send(control.ProcessQueryReply{ProcessList: batch})
	}
	return nil
}

// sovControl is imported from control.pb.go and is used to calculate the
// ProcessSummary size, that we need to evaluate while building the entries
// list
func sovControl(x uint64) (n int) {
	return (math_bits.Len64(x|1) + 6) / 7
}

func (s processServer) QuerySummary(
	ctx context.Context, req control.ProcessQueryRequest,
	send func(control.ProcessSummary),
) error {
	s.TaskAdd()
	defer s.TaskDone()

	summary, err := s.o.ProcessStorage.QuerySummary(req.Filter)
	if err != nil {
		return err
	}
	var entries []*control.ProcessSummary_Entry
	if len(summary) == 0 {
		return nil
	}
	var (
		maxSize = 1024 * 512
		size    = 0
	)
	for _, entry := range summary {
		entry := control.ProcessSummary_Entry{
			PipelineID:  entry.PipelineID,
			EmitterID:   entry.EmitterID,
			Status:      entry.Status,
			ResultAcked: entry.ResultAcked,
			Count:       entry.Count,
		}
		entrySize := entry.Size()
		entrySizeDelta := entrySize + 1 + sovControl(uint64(entrySize))
		if size+entrySizeDelta > maxSize {
			send(control.ProcessSummary{Entries: entries})
			entries = entries[0:0]
			size = 0
		}
		entries = append(entries, &entry)
		size += entrySizeDelta
	}
	if len(entries) != 0 {
		send(control.ProcessSummary{Entries: entries})
	}

	return nil
}

func (s processServer) Control(
	ctx context.Context, req control.ProcessControlRequest,
) (err error) {
	s.TaskAdd()
	defer s.TaskDone()
	p, err := s.o.ProcessManager.Get(req.ProcessID)
	if err != nil {
		return err
	}
	switch req.Command {
	case control.ProcessControlRequest_PAUSE:
		err = p.Pause()
	case control.ProcessControlRequest_RESUME:
		err = p.Resume()
	case control.ProcessControlRequest_CANCEL:
		err = p.Cancel("Canceled by user")
	default:
		err = fmt.Errorf("Invalid command: %s", err)
	}
	return
}

func (s processServer) Export(
	ctx context.Context, req control.ProcessExportRequest,
) (
	resp control.ProcessExportReply, err error,
) {
	s.TaskAdd()
	defer s.TaskDone()

	for _, pid := range req.ProcessIDs {
		p, err := s.o.ProcessManager.Get(pid)
		if err != nil {
			panic(err)
		}

		export, err := p.Export(s.o.EnvelopeStorage, s.Log)
		if err != nil {
			panic(err)
		}

		b, err := yaml.Marshal(export)
		if err != nil {
			panic(err)
		}

		resp.ProcessList = append(resp.ProcessList, string(b))
	}
	return
}

func (s processServer) ExportStream(
	ctx context.Context, req control.ProcessExportRequest,
	send func(control.OctetStreamList),
) (
	err error,
) {
	s.TaskAdd()
	defer s.TaskDone()

	for index, pid := range req.ProcessIDs {
		if ctx.Err() != nil {
			return ctx.Err()
		}
		p := s.o.ProcessManager.GetNoErr(pid)
		if p == nil {
			continue
		}

		wg := sync.WaitGroup{}
		wg.Add(1)

		r, w := io.Pipe()
		defer w.Close()

		go func() {
			defer wg.Done()
			var (
				bufSize = 1024 * 768
				minSize = 1024 * 512
				pos     int
			)

			buffer := make([]byte, bufSize)

			for {
				n, err := r.Read(buffer[pos:])
				pos += n
				if pos > 0 && (pos >= minSize || err == io.EOF) {
					if ctx.Err() == nil {
						send(control.OctetStreamList{
							Progression:    int64(index),
							MaxProgression: int64(len(req.ProcessIDs)),
							Chunks: []control.OctetStreamList_Chunk{
								{Data: buffer[:pos],
									Index: int64(index),
								}},
						})
					}
					pos = 0
				}
				if err == io.EOF {
					return
				}
			}
		}()

		if err := p.ExportStream(
			ctx,
			s.o.EnvelopeStorage,
			s.o.JobStorage,
			s.o.LogStorage,
			s.Log, w,
			req.ExportFull || req.ExportEnvelope,
			req.ExportFull || req.ExportJobs,
			req.ExportFull || req.ExportLogs,
			req.ExportFull || req.ExportProcessLogs,
			req.EnvelopeSizeLimit,
			req.LogLevel,
		); err != nil {
			panic(err)
		}
		w.Close()
		wg.Wait()

	}
	return
}

func (s processServer) Purge(
	ctx context.Context, req control.ProcessPurgeRequest,
	send func(control.TaskProgress),
) error {
	s.TaskAdd()
	defer s.TaskDone()

	if len(req.ProcessIDs) == 0 {
		send(control.TaskProgress{})
		return nil
	}

	var wg sync.WaitGroup
	jobCount := len(req.ProcessIDs) / s.o.PurgeBatchSize
	if jobCount == 0 {
		jobCount = 1
	} else if jobCount > s.o.MaxPurgeJobs {
		jobCount = s.o.MaxPurgeJobs
	}

	type res struct {
		err error
		ids []api.UUID
	}

	jobsCh := make(chan []api.UUID)
	resCh := make(chan res, jobCount)
	stopCh := make(chan struct{})

	wg.Add(jobCount + 1)

	// Spin up jobs
	for i := 0; i != jobCount; i++ {
		go func() {
			defer wg.Done()
			for {
				select {
				case batch, ok := <-jobsCh:
					if !ok {
						return
					}
					ids, err := s.o.ProcessManager.PurgeMany(batch)
					if err == nil {
						err = s.o.JobStorage.Purge(ids)
					}
					resCh <- res{err, ids}
				case <-ctx.Done():
					return
				}
			}
		}()
	}

	var (
		max     = uint32(len(req.ProcessIDs))
		counter uint32
	)

	// Push jobs to the queue
	go func() {
		defer wg.Done()
		defer close(jobsCh)
		var batches [][]api.UUID

		ids := req.ProcessIDs
		for s.o.PurgeBatchSize < len(ids) {
			ids, batches = ids[s.o.PurgeBatchSize:], append(batches, ids[0:s.o.PurgeBatchSize:s.o.PurgeBatchSize])
		}
		batches = append(batches, ids)
		for _, batch := range batches {
			select {
			case <-stopCh:
				return
			case <-ctx.Done():
				return
			case jobsCh <- batch:
			}
		}
	}()

	// Receive results and report back to the caller
	for {
		select {
		case <-ctx.Done():
			wg.Wait()
			return nil

		case r := <-resCh:
			if r.err != nil {
				stopCh <- struct{}{}
				wg.Wait()
				return r.err
			}
			counter += uint32(len(r.ids))
			send(control.TaskProgress{Progression: counter, MaxProgression: max})
			if counter == max {
				wg.Wait()
				return nil
			}
		}
	}
}

func (s processServer) Replay(
	ctx context.Context, req control.ProcessReplayRequest,
) (api.Process, error) {
	process, err := s.o.ProcessManager.Get(req.Process.ID)
	if err != nil {
		return api.Process{}, fmt.Errorf("Could not get process: %s", err)
	}
	newProcess, err := s.o.ProcessManager.Replay(process, s.o.EnvelopeStorage, req.Match)
	if err != nil {
		return api.Process{}, fmt.Errorf("Could not replay process: %s", err)
	}
	state := s.o.ProcessStorage.GetState(newProcess.ID)
	b, err := json.Marshal(state)
	if err != nil {
		panic(err)
	}
	createdAt := newProcess.CreatedAt
	return api.Process{
		ID:                newProcess.ID,
		TriggerEmitterID:  newProcess.TriggerEmitterID,
		TriggerEnvelopeID: newProcess.TriggerEnvelopeID,
		CreatedAt:         &createdAt,
		PipelineID:        newProcess.Pipeline.ID,
		GroupID:           newProcess.GroupID,
		ReplayOf:          newProcess.ReplayOf,
		Status:            state.Status,
		State:             string(b),
	}, nil
}
