package metrics

import (
	"context"
	"sync"
	"sync/atomic"
	"time"

	"github.com/prometheus/client_golang/prometheus"
	"github.com/rs/zerolog"
)

func NewSlowCollector(
	ctx context.Context, name string, c prometheus.Collector,
	delayHint time.Duration, log zerolog.Logger,
) *SlowCollector {
	sc := SlowCollector{
		Collector: c,
		metricsC:  make(chan []prometheus.Metric, 1),
		delayHint: delayHint,
		log:       log.With().Str("slow-collector", name).Logger(),
	}
	sc.ctx, sc.stop = context.WithCancel(ctx)
	sc.wg.Add(1)
	go sc.run()
	return &sc
}

type SlowCollector struct {
	ctx       context.Context
	Collector prometheus.Collector

	stop      func()
	wg        sync.WaitGroup
	metricsC  chan []prometheus.Metric
	delayHint time.Duration
	log       zerolog.Logger
}

// Describe implements the prometheus.Collector interface.
func (c *SlowCollector) Describe(ch chan<- *prometheus.Desc) {
	c.Collector.Describe(ch)
}

// Collect implements the prometheus.Collector interface.
func (c *SlowCollector) Collect(ch chan<- prometheus.Metric) {
	select {
	case metrics := <-c.metricsC:
		for _, m := range metrics {
			ch <- m
		}
	case <-time.After(100 * time.Millisecond):
		c.log.Warn().Msg("could not get metrics fast enough. This is harmless on startup but should not occur shortly after")
	}
}

func (c *SlowCollector) Close() {
	c.stop()
	c.wg.Wait()
}

func (c *SlowCollector) run() {
	defer c.wg.Done()
	defer close(c.metricsC)

	var (
		delayHint = c.delayHint
		buffer    = make(chan []prometheus.Metric)

		lastDuration  time.Duration
		lastCollectCh = make(chan time.Time, 1)

		wg sync.WaitGroup
	)

	defer wg.Wait()

	defer close(buffer)

	wg.Add(1)

	go func() {
		defer wg.Done()

		var (
			cont    = true
			metrics = <-buffer
		)

		for cont {
			select {
			case c.metricsC <- metrics:
				select {
				case <-lastCollectCh:
				default:
				}
				lastCollectCh <- time.Now()
			case metrics, cont = <-buffer:
			}
		}
	}()

	doCollect := func() []prometheus.Metric {
		var (
			metrics  []prometheus.Metric
			resultCh = make(chan prometheus.Metric)
		)

		c.log.Info().Msg("collecting")

		go func() {
			start := time.Now()

			c.Collector.Collect(resultCh)
			atomic.StoreInt64((*int64)(&lastDuration), int64(time.Since(start)))

			close(resultCh)
		}()

		for m := range resultCh {
			metrics = append(metrics, m)
		}
		return metrics
	}

	buffer <- doCollect()

	const defaultCollectDelay = time.Minute

	var (
		nextCollect        = time.Now().Add(defaultCollectDelay)
		delayTimer         = time.NewTimer(time.Until(nextCollect))
		previousCollect    time.Time
		delayHintCandidate = delayHint
	)

	for {
		select {
		case lastCollect := <-lastCollectCh:
			if !previousCollect.IsZero() {
				// auto-adjust delayHint if needed
				delay := lastCollect.Sub(previousCollect).Round(time.Second)
				if delay == 0 {
					delay = time.Second
				}
				if delayHintCandidate == delay && delay != delayHint {
					delayHint = delay
					c.log.Info().Dur("delay-hint", delayHint).Msg("New delay hint")
				}
				delayHintCandidate = delay
			}
			previousCollect = lastCollect

			newNextCollect := lastCollect.Add(
				delayHint -
					(time.Duration(atomic.LoadInt64((*int64)(&lastDuration))) * 11 / 10))

			if newNextCollect.Before(nextCollect) {
				delay := time.Until(newNextCollect)
				if delay <= 0 {
					delay = time.Millisecond
				}

				if !delayTimer.Stop() {
					<-delayTimer.C
				}

				nextCollect = newNextCollect

				delayTimer.Reset(delay)
			}

		case <-delayTimer.C:
			buffer <- doCollect()

			nextCollect = time.Now().Add(defaultCollectDelay)
			delayTimer.Reset(time.Until(nextCollect))

		case <-c.ctx.Done():
			return
		}
	}
}
