package metrics

import (
	"context"
	"fmt"

	"github.com/prometheus/client_golang/prometheus"
	"github.com/rs/zerolog"
	"xbus.io/go-xbus/v3/api"
	"xbus.io/go-xbus/v3/api/control"
	"xbus.io/xbus/services/tools"
)

// NewSummaryCollector creates a new SummaryCollector
func NewSummaryCollector(
	control *control.Client, actors *tools.ActorCache, pipelines *tools.PipelineCache,
	log zerolog.Logger,
) *SummaryCollector {
	c := SummaryCollector{
		log:       log,
		control:   control,
		actors:    actors,
		pipelines: pipelines,

		processCountDesc: prometheus.NewDesc(
			prometheus.BuildFQName("xbus", "process", "count"),
			"Number of process in the bus",
			[]string{"pipeline", "pipeline_version", "emitter", "status", "result_acked"},
			nil,
		),

		storageCountDesc: prometheus.NewDesc(
			prometheus.BuildFQName("xbus", "storage", "count"),
			"Number of items in the storage",
			[]string{"name", "implementation"},
			prometheus.Labels{"type": "process"},
		),
	}
	return &c
}

// SummaryCollector implements the prometheus.Collector interface
type SummaryCollector struct {
	log zerolog.Logger

	control   *control.Client
	actors    *tools.ActorCache
	pipelines *tools.PipelineCache

	storageCountDesc *prometheus.Desc
	processCountDesc *prometheus.Desc
}

// Describe implements the prometheus.Collector interface.
func (c SummaryCollector) Describe(ch chan<- *prometheus.Desc) {
	ch <- c.processCountDesc
}

func (c *SummaryCollector) getActor(id api.UUID) string {
	a, err := c.actors.Get(id)
	if err != nil {
		return id.String()
	}
	return a.Name
}

func (c *SummaryCollector) getPipeline(id api.UUID) (string, string) {
	p, err := c.pipelines.Get(id)
	if err != nil {
		return id.String(), ""
	}
	return p.Name, p.Version
}

// Collect implements the prometheus.Collector interface.
func (c *SummaryCollector) Collect(ch chan<- prometheus.Metric) {
	var (
		summary control.ProcessSummary
	)

	if err := c.control.Process.QuerySummary(
		context.Background(), control.ProcessQueryRequest{},
		func(ctx context.Context, r control.ProcessSummary) {
			summary.Entries = append(summary.Entries, r.Entries...)
		},
	); err != nil {
		c.log.Err(err).
			Str("collector", "process-summary").
			Msg("failed to query the process summary")
		return
	}

	var count uint64

	for _, entry := range summary.Entries {
		count += entry.Count
		emitter := c.getActor(entry.EmitterID)
		pipeline, version := c.getPipeline(entry.PipelineID)
		ch <- prometheus.MustNewConstMetric(
			c.processCountDesc,
			prometheus.GaugeValue,
			float64(entry.Count),
			pipeline,
			version,
			emitter,
			entry.Status.String(),
			fmt.Sprintf("%v", entry.ResultAcked),
		)
	}

	ch <- prometheus.MustNewConstMetric(
		c.storageCountDesc,
		prometheus.GaugeValue,
		float64(count),
		"default",
		"default",
	)
}
